@@ -24,18 +24,15 @@
 # CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
 
-
 from unicodedata import normalize
 
-from version import *
-import bitcoin
-from bitcoin import pw_encode, pw_decode, bip32_root, bip32_private_derivation, bip32_public_derivation, bip32_private_key, deserialize_xkey
-from bitcoin import public_key_from_private_key, public_key_to_bc_address
-from bitcoin import *
-
-from bitcoin import is_old_seed, is_new_seed, is_seed
-from util import PrintError, InvalidPassword
-from mnemonic import Mnemonic
+from . import bitcoin
+from .bitcoin import *
+from . import constants
+from .util import (PrintError, InvalidPassword, hfu, WalletFileException,
+                   BitcoinException)
+from .mnemonic import Mnemonic, load_wordlist
+from .plugins import run_hook
 
 
 class KeyStore(PrintError):
@@ -49,6 +46,10 @@ class KeyStore(PrintError):
     def can_import(self):
         return False
 
+    def may_have_password(self):
+        """Returns whether the keystore can be encrypted with a password."""
+        raise NotImplementedError()
+
     def get_tx_derivations(self, tx):
         keypairs = {}
         for txin in tx.inputs():
@@ -56,7 +57,7 @@ class KeyStore(PrintError):
             if num_sig is None:
                 continue
             x_signatures = txin['signatures']
-            signatures = filter(None, x_signatures)
+            signatures = [sig for sig in x_signatures if sig]
             if len(signatures) == num_sig:
                 # input is complete
                 continue
@@ -75,6 +76,9 @@ class KeyStore(PrintError):
             return False
         return bool(self.get_tx_derivations(tx))
 
+    def ready_to_sign(self):
+        return not self.is_watching_only()
+
 
 class Software_KeyStore(KeyStore):
 
@@ -85,14 +89,13 @@ class Software_KeyStore(KeyStore):
         return not self.is_watching_only()
 
     def sign_message(self, sequence, message, password):
-        sec = self.get_private_key(sequence, password)
-        key = regenerate_key(sec)
-        compressed = is_compressed(sec)
+        privkey, compressed = self.get_private_key(sequence, password)
+        key = regenerate_key(privkey)
         return key.sign_message(message, compressed)
 
     def decrypt_message(self, sequence, message, password):
-        sec = self.get_private_key(sequence, password)
-        ec = regenerate_key(sec)
+        privkey, compressed = self.get_private_key(sequence, password)
+        ec = regenerate_key(privkey)
         decrypted = ec.decrypt_message(message)
         return decrypted
 
@@ -116,15 +119,10 @@ class Imported_KeyStore(Software_KeyStore):
     def __init__(self, d):
         Software_KeyStore.__init__(self)
         self.keypairs = d.get('keypairs', {})
-        self.receiving_pubkeys = self.keypairs.keys()
-        self.change_pubkeys = []
 
     def is_deterministic(self):
         return False
 
-    def can_change_password(self):
-        return True
-
     def get_master_public_key(self):
         return None
 
@@ -138,53 +136,41 @@ class Imported_KeyStore(Software_KeyStore):
         return True
 
     def check_password(self, password):
-        self.get_private_key((0,0), password)
+        pubkey = list(self.keypairs.keys())[0]
+        self.get_private_key(pubkey, password)
 
-    def import_key(self, sec, password):
-        if not self.can_import():
-            raise BaseException('This wallet cannot import private keys')
-        try:
-            pubkey = public_key_from_private_key(sec)
-        except Exception:
-            raise BaseException('Invalid private key')
-        if pubkey in self.keypairs:
-            raise BaseException('Private key already in keystore')
-        self.keypairs[pubkey] = sec
-        self.receiving_pubkeys = self.keypairs.keys()
-        return pubkey
+    def import_privkey(self, sec, password):
+        txin_type, privkey, compressed = deserialize_privkey(sec)
+        pubkey = public_key_from_private_key(privkey, compressed)
+        # re-serialize the key so the internal storage format is consistent
+        serialized_privkey = serialize_privkey(
+            privkey, compressed, txin_type, internal_use=True)
+        # NOTE: if the same pubkey is reused for multiple addresses (script types),
+        # there will only be one pubkey-privkey pair for it in self.keypairs,
+        # and the privkey will encode a txin_type but that txin_type cannot be trusted.
+        # Removing keys complicates this further.
+        self.keypairs[pubkey] = pw_encode(serialized_privkey, password)
+        return txin_type, pubkey
 
     def delete_imported_key(self, key):
         self.keypairs.pop(key)
 
-    def get_public_key(self, sequence):
-        for_change, i = sequence
-        pubkey = (self.change_pubkeys if for_change else self.receiving_pubkeys)[i]
-        return pubkey
-
-    def get_xpubkey(self, c, i):
-        return self.get_public_key((c,i))
-
-    def get_private_key(self, sequence, password):
-        for_change, i = sequence
-        assert for_change == 0
-        pubkey = self.receiving_pubkeys[i]
-        pk = pw_decode(self.keypairs[pubkey], password)
+    def get_private_key(self, pubkey, password):
+        sec = pw_decode(self.keypairs[pubkey], password)
+        txin_type, privkey, compressed = deserialize_privkey(sec)
         # this checks the password
-        if pubkey != public_key_from_private_key(pk):
+        if pubkey != public_key_from_private_key(privkey, compressed):
             raise InvalidPassword()
-        return pk
+        return privkey, compressed
 
     def get_pubkey_derivation(self, x_pubkey):
         if x_pubkey[0:2] in ['02', '03', '04']:
-            if x_pubkey in self.receiving_pubkeys:
-                i = self.receiving_pubkeys.index(x_pubkey)
-                return (False, i)
+            if x_pubkey in self.keypairs.keys():
+                return x_pubkey
         elif x_pubkey[0:2] == 'fd':
-            # fixme: this assumes p2pkh
-            _, addr = xpubkey_to_address(x_pubkey)
-            for i, pubkey in enumerate(self.receiving_pubkeys):
-                if public_key_to_bc_address(pubkey.decode('hex')) == addr:
-                    return (False, i)
+            addr = bitcoin.script_to_address(x_pubkey[2:])
+            if addr in self.addresses:
+                return self.addresses[addr].get('pubkey')
 
     def update_password(self, old_password, new_password):
         self.check_password(old_password)
@@ -196,6 +182,7 @@ class Imported_KeyStore(Software_KeyStore):
             self.keypairs[k] = c
 
 
+
 class Deterministic_KeyStore(Software_KeyStore):
 
     def __init__(self, d):
@@ -220,20 +207,16 @@ class Deterministic_KeyStore(Software_KeyStore):
     def is_watching_only(self):
         return not self.has_seed()
 
-    def can_change_password(self):
-        return not self.is_watching_only()
-
     def add_seed(self, seed):
         if self.seed:
             raise Exception("a seed exists")
         self.seed = self.format_seed(seed)
 
     def get_seed(self, password):
-        return pw_decode(self.seed, password).encode('utf8')
+        return pw_decode(self.seed, password)
 
     def get_passphrase(self, password):
-        return pw_decode(self.passphrase, password).encode('utf8') if self.passphrase else ''
-
+        return pw_decode(self.passphrase, password) if self.passphrase else ''
 
 
 class Xpub:
@@ -258,25 +241,25 @@ class Xpub:
 
     @classmethod
     def get_pubkey_from_xpub(self, xpub, sequence):
-        _, _, _, c, cK = deserialize_xkey(xpub)
+        _, _, _, _, c, cK = deserialize_xpub(xpub)
         for i in sequence:
             cK, c = CKD_pub(cK, c, i)
-        return cK.encode('hex')
+        return bh2u(cK)
 
     def get_xpubkey(self, c, i):
         s = ''.join(map(lambda x: bitcoin.int_to_hex(x,2), (c, i)))
-        return 'ff' + bitcoin.DecodeBase58Check(self.xpub).encode('hex') + s
+        return 'ff' + bh2u(bitcoin.DecodeBase58Check(self.xpub)) + s
 
     @classmethod
     def parse_xpubkey(self, pubkey):
         assert pubkey[0:2] == 'ff'
-        pk = pubkey.decode('hex')
+        pk = bfh(pubkey)
         pk = pk[1:]
         xkey = bitcoin.EncodeBase58Check(pk[0:78])
         dd = pk[78:]
         s = []
         while dd:
-            n = int(bitcoin.rev_hex(dd[0:2].encode('hex')), 16)
+            n = int(bitcoin.rev_hex(bh2u(dd[0:2])), 16)
             dd = dd[2:]
             s.append(n)
         assert len(s) == 2
@@ -291,7 +274,6 @@ class Xpub:
         return derivation
 
 
-
 class BIP32_KeyStore(Deterministic_KeyStore, Xpub):
 
     def __init__(self, d):
@@ -315,7 +297,7 @@ class BIP32_KeyStore(Deterministic_KeyStore, Xpub):
 
     def check_password(self, password):
         xprv = pw_decode(self.xprv, password)
-        if deserialize_xkey(xprv)[3] != deserialize_xkey(self.xpub)[3]:
+        if deserialize_xprv(xprv)[4] != deserialize_xpub(self.xpub)[4]:
             raise InvalidPassword()
 
     def update_password(self, old_password, new_password):
@@ -335,23 +317,20 @@ class BIP32_KeyStore(Deterministic_KeyStore, Xpub):
     def is_watching_only(self):
         return self.xprv is None
 
-    def get_mnemonic(self, password):
-        return self.get_seed(password)
-
     def add_xprv(self, xprv):
         self.xprv = xprv
         self.xpub = bitcoin.xpub_from_xprv(xprv)
 
-    def add_xprv_from_seed(self, bip32_seed, derivation):
-        xprv, xpub = bip32_root(bip32_seed)
+    def add_xprv_from_seed(self, bip32_seed, xtype, derivation):
+        xprv, xpub = bip32_root(bip32_seed, xtype)
         xprv, xpub = bip32_private_derivation(xprv, "m/", derivation)
         self.add_xprv(xprv)
 
     def get_private_key(self, sequence, password):
         xprv = self.get_master_private_key(password)
-        _, _, _, c, k = deserialize_xkey(xprv)
+        _, _, _, _, c, k = deserialize_xprv(xprv)
         pk = bip32_private_key(sequence, k, c)
-        return pk
+        return pk, True
 
 
 
@@ -361,6 +340,9 @@ class Old_KeyStore(Deterministic_KeyStore):
         Deterministic_KeyStore.__init__(self, d)
         self.mpk = d.get('mpk')
 
+    def get_hex_seed(self, password):
+        return pw_decode(self.seed, password).encode('utf8')
+
     def dump(self):
         d = Deterministic_KeyStore.dump(self)
         d['mpk'] = self.mpk
@@ -369,18 +351,19 @@ class Old_KeyStore(Deterministic_KeyStore):
 
     def add_seed(self, seedphrase):
         Deterministic_KeyStore.add_seed(self, seedphrase)
-        self.mpk = self.mpk_from_seed(self.seed)
+        s = self.get_hex_seed(None)
+        self.mpk = self.mpk_from_seed(s)
 
     def add_master_public_key(self, mpk):
         self.mpk = mpk
 
     def format_seed(self, seed):
-        import old_mnemonic
+        from . import old_mnemonic, mnemonic
+        seed = mnemonic.normalize_text(seed)
         # see if seed was entered as hex
-        seed = seed.strip()
         if seed:
             try:
-                seed.decode('hex')
+                bfh(seed)
                 return str(seed)
             except Exception:
                 pass
@@ -390,9 +373,9 @@ class Old_KeyStore(Deterministic_KeyStore):
             raise Exception("Invalid seed")
         return seed
 
-    def get_mnemonic(self, password):
-        import old_mnemonic
-        s = self.get_seed(password)
+    def get_seed(self, password):
+        from . import old_mnemonic
+        s = self.get_hex_seed(password)
         return ' '.join(old_mnemonic.mn_encode(s))
 
     @classmethod
@@ -400,7 +383,7 @@ class Old_KeyStore(Deterministic_KeyStore):
         secexp = klass.stretch_key(seed)
         master_private_key = ecdsa.SigningKey.from_secret_exponent(secexp, curve = SECP256k1)
         master_public_key = master_private_key.get_verifying_key().to_string()
-        return master_public_key.encode('hex')
+        return bh2u(master_public_key)
 
     @classmethod
     def stretch_key(self, seed):
@@ -411,20 +394,15 @@ class Old_KeyStore(Deterministic_KeyStore):
 
     @classmethod
     def get_sequence(self, mpk, for_change, n):
-        return string_to_number(Hash("%d:%d:"%(n, for_change) + mpk.decode('hex')))
-
-    def get_address(self, for_change, n):
-        pubkey = self.get_pubkey(for_change, n)
-        address = public_key_to_bc_address(pubkey.decode('hex'))
-        return address
+        return string_to_number(Hash(("%d:%d:"%(n, for_change)).encode('ascii') + bfh(mpk)))
 
     @classmethod
     def get_pubkey_from_mpk(self, mpk, for_change, n):
         z = self.get_sequence(mpk, for_change, n)
-        master_public_key = ecdsa.VerifyingKey.from_string(mpk.decode('hex'), curve = SECP256k1)
+        master_public_key = ecdsa.VerifyingKey.from_string(bfh(mpk), curve = SECP256k1)
         pubkey_point = master_public_key.pubkey.point + z*SECP256k1.generator
         public_key2 = ecdsa.VerifyingKey.from_public_point(pubkey_point, curve = SECP256k1)
-        return '04' + public_key2.to_string().encode('hex')
+        return '04' + bh2u(public_key2.to_string())
 
     def derive_pubkey(self, for_change, n):
         return self.get_pubkey_from_mpk(self.mpk, for_change, n)
@@ -433,27 +411,26 @@ class Old_KeyStore(Deterministic_KeyStore):
         order = generator_secp256k1.order()
         secexp = (secexp + self.get_sequence(self.mpk, for_change, n)) % order
         pk = number_to_string(secexp, generator_secp256k1.order())
-        compressed = False
-        return SecretToASecret(pk, compressed)
+        return pk
 
     def get_private_key(self, sequence, password):
-        seed = self.get_seed(password)
+        seed = self.get_hex_seed(password)
         self.check_seed(seed)
         for_change, n = sequence
         secexp = self.stretch_key(seed)
         pk = self.get_private_key_from_stretched_exponent(for_change, n, secexp)
-        return pk
+        return pk, False
 
     def check_seed(self, seed):
         secexp = self.stretch_key(seed)
         master_private_key = ecdsa.SigningKey.from_secret_exponent( secexp, curve = SECP256k1 )
         master_public_key = master_private_key.get_verifying_key().to_string()
-        if master_public_key != self.mpk.decode('hex'):
-            print_error('invalid password (mpk)', self.mpk, master_public_key.encode('hex'))
+        if master_public_key != bfh(self.mpk):
+            print_error('invalid password (mpk)', self.mpk, bh2u(master_public_key))
             raise InvalidPassword()
 
     def check_password(self, password):
-        seed = self.get_seed(password)
+        seed = self.get_hex_seed(password)
         self.check_seed(seed)
 
     def get_master_public_key(self):
@@ -490,10 +467,11 @@ class Old_KeyStore(Deterministic_KeyStore):
         if new_password == '':
             new_password = None
         if self.has_seed():
-            decoded = self.get_seed(old_password)
+            decoded = pw_decode(self.seed, old_password)
             self.seed = pw_encode(decoded, new_password)
 
 
+
 class Hardware_KeyStore(KeyStore, Xpub):
     # Derived classes must set:
     #   - device
@@ -513,6 +491,7 @@ class Hardware_KeyStore(KeyStore, Xpub):
         self.label = d.get('label')
         self.derivation = d.get('derivation')
         self.handler = None
+        run_hook('init_keystore', self)
 
     def set_label(self, label):
         self.label = label
@@ -533,7 +512,7 @@ class Hardware_KeyStore(KeyStore, Xpub):
         }
 
     def unpaired(self):
-        '''A device paired with the wallet was diconnected.  This can be
+        '''A device paired with the wallet was disconnected.  This can be
         called in any thread context.'''
         self.print_error("unpaired")
 
@@ -551,13 +530,28 @@ class Hardware_KeyStore(KeyStore, Xpub):
         assert not self.has_seed()
         return False
 
-    def can_change_password(self):
-        return False
+    def get_password_for_storage_encryption(self):
+        from .storage import get_derivation_used_for_hw_device_encryption
+        client = self.plugin.get_client(self)
+        derivation = get_derivation_used_for_hw_device_encryption()
+        xpub = client.get_xpub(derivation, "standard")
+        password = self.get_pubkey_from_xpub(xpub, ())
+        return password
 
+    def has_usable_connection_with_device(self):
+        if not hasattr(self, 'plugin'):
+            return False
+        client = self.plugin.get_client(self, force_pair=False)
+        if client is None:
+            return False
+        return client.has_usable_connection_with_device()
+
+    def ready_to_sign(self):
+        return super().ready_to_sign() and self.has_usable_connection_with_device()
 
 
 def bip39_normalize_passphrase(passphrase):
-    return normalize('NFKD', unicode(passphrase or ''))
+    return normalize('NFKD', passphrase or '')
 
 def bip39_to_seed(mnemonic, passphrase):
     import pbkdf2, hashlib, hmac
@@ -568,6 +562,51 @@ def bip39_to_seed(mnemonic, passphrase):
                          iterations = PBKDF2_ROUNDS, macmodule = hmac,
                          digestmodule = hashlib.sha512).read(64)
 
+# returns tuple (is_checksum_valid, is_wordlist_valid)
+def bip39_is_checksum_valid(mnemonic):
+    words = [ normalize('NFKD', word) for word in mnemonic.split() ]
+    words_len = len(words)
+    wordlist = load_wordlist("english.txt")
+    n = len(wordlist)
+    checksum_length = 11*words_len//33
+    entropy_length = 32*checksum_length
+    i = 0
+    words.reverse()
+    while words:
+        w = words.pop()
+        try:
+            k = wordlist.index(w)
+        except ValueError:
+            return False, False
+        i = i*n + k
+    if words_len not in [12, 15, 18, 21, 24]:
+        return False, True
+    entropy = i >> checksum_length
+    checksum = i % 2**checksum_length
+    h = '{:x}'.format(entropy)
+    while len(h) < entropy_length/4:
+        h = '0'+h
+    b = bytearray.fromhex(h)
+    hashed = int(hfu(hashlib.sha256(b).digest()), 16)
+    calculated_checksum = hashed >> (256 - checksum_length)
+    return checksum == calculated_checksum, True
+
+def from_bip39_seed(seed, passphrase, derivation):
+    k = BIP32_KeyStore({})
+    bip32_seed = bip39_to_seed(seed, passphrase)
+    xtype = xtype_from_derivation(derivation)
+    k.add_xprv_from_seed(bip32_seed, xtype, derivation)
+    return k
+
+
+def xtype_from_derivation(derivation):
+    """Returns the script type to be used for this derivation."""
+    if derivation.startswith("m/84'"):
+        return 'p2wpkh'
+    elif derivation.startswith("m/49'"):
+        return 'p2wpkh-p2sh'
+    else:
+        return 'standard'
 
 
 # extended pubkeys
@@ -575,12 +614,17 @@ def bip39_to_seed(mnemonic, passphrase):
 def is_xpubkey(x_pubkey):
     return x_pubkey[0:2] == 'ff'
 
+
 def parse_xpubkey(x_pubkey):
     assert x_pubkey[0:2] == 'ff'
     return BIP32_KeyStore.parse_xpubkey(x_pubkey)
 
+
 def xpubkey_to_address(x_pubkey):
-    if x_pubkey[0:2] in ['02','03','04']:
+    if x_pubkey[0:2] == 'fd':
+        address = bitcoin.script_to_address(x_pubkey[2:])
+        return x_pubkey, address
+    if x_pubkey[0:2] in ['02', '03', '04']:
         pubkey = x_pubkey
     elif x_pubkey[0:2] == 'ff':
         xpub, s = BIP32_KeyStore.parse_xpubkey(x_pubkey)
@@ -588,17 +632,16 @@ def xpubkey_to_address(x_pubkey):
     elif x_pubkey[0:2] == 'fe':
         mpk, s = Old_KeyStore.parse_xpubkey(x_pubkey)
         pubkey = Old_KeyStore.get_pubkey_from_mpk(mpk, s[0], s[1])
-    elif x_pubkey[0:2] == 'fd':
-        addrtype = ord(x_pubkey[2:4].decode('hex'))
-        hash160 = x_pubkey[4:].decode('hex')
-        pubkey = None
-        address = hash_160_to_bc_address(hash160, addrtype)
     else:
-        raise BaseException("Cannnot parse pubkey")
+        raise BitcoinException("Cannot parse pubkey. prefix: {}"
+                               .format(x_pubkey[0:2]))
     if pubkey:
-        address = public_key_to_bc_address(pubkey.decode('hex'))
+        address = public_key_to_p2pkh(bfh(pubkey))
     return pubkey, address
 
+def xpubkey_to_pubkey(x_pubkey):
+    pubkey, address = xpubkey_to_address(x_pubkey)
+    return pubkey
 
 hw_keystores = {}
 
@@ -610,14 +653,15 @@ def hardware_keystore(d):
     if hw_type in hw_keystores:
         constructor = hw_keystores[hw_type]
         return constructor(d)
-    raise BaseException('unknown hardware type', hw_type)
+    raise WalletFileException('unknown hardware type: {}'.format(hw_type))
 
 def load_keystore(storage, name):
-    w = storage.get('wallet_type', 'standard')
     d = storage.get(name, {})
     t = d.get('type')
     if not t:
-        raise BaseException('wallet format requires update')
+        raise WalletFileException(
+            'Wallet format requires update.\n'
+            'Cannot find keystore for name {}'.format(name))
     if t == 'old':
         k = Old_KeyStore(d)
     elif t == 'imported':
@@ -627,11 +671,11 @@ def load_keystore(storage, name):
     elif t == 'hardware':
         k = hardware_keystore(d)
     else:
-        raise BaseException('unknown wallet type', t)
+        raise WalletFileException(
+            'Unknown type {} for keystore named {}'.format(t, name))
     return k
 
 
-
 def is_old_mpk(mpk):
     try:
         int(mpk, 16)
@@ -639,57 +683,59 @@ def is_old_mpk(mpk):
         return False
     return len(mpk) == 128
 
-def is_xpub(text):
-    if text[0:4] != 'xpub':
-        return False
-    try:
-        deserialize_xkey(text)
-        return True
-    except:
-        return False
-
-def is_xprv(text):
-    if text[0:4] != 'xprv':
-        return False
-    try:
-        deserialize_xkey(text)
-        return True
-    except:
-        return False
 
 def is_address_list(text):
     parts = text.split()
     return bool(parts) and all(bitcoin.is_address(x) for x in parts)
 
+
+def get_private_keys(text):
+    parts = text.split('\n')
+    parts = map(lambda x: ''.join(x.split()), parts)
+    parts = list(filter(bool, parts))
+    if bool(parts) and all(bitcoin.is_private_key(x) for x in parts):
+        return parts
+
+
 def is_private_key_list(text):
-    parts = text.split()
-    return bool(parts) and all(bitcoin.is_private_key(x) for x in parts)
+    return bool(get_private_keys(text))
 
 
 is_mpk = lambda x: is_old_mpk(x) or is_xpub(x)
 is_private = lambda x: is_seed(x) or is_xprv(x) or is_private_key_list(x)
-is_any_key = lambda x: is_old_mpk(x) or is_xprv(x) or is_xpub(x) or is_address_list(x) or is_private_key_list(x)
+is_master_key = lambda x: is_old_mpk(x) or is_xprv(x) or is_xpub(x)
 is_private_key = lambda x: is_xprv(x) or is_private_key_list(x)
 is_bip32_key = lambda x: is_xprv(x) or is_xpub(x)
 
-def bip44_derivation(account_id):
-    return "m/44'/0'/%d'"% int(account_id)
 
-def from_seed(seed, passphrase):
-    if is_old_seed(seed):
+def bip44_derivation(account_id, bip43_purpose=44):
+    coin = 1 if constants.net.TESTNET else 0
+    return "m/%d'/%d'/%d'" % (bip43_purpose, coin, int(account_id))
+
+def from_seed(seed, passphrase, is_p2sh):
+    t = seed_type(seed)
+    if t == 'old':
         keystore = Old_KeyStore({})
         keystore.add_seed(seed)
-    elif is_new_seed(seed):
+    elif t in ['standard', 'segwit']:
         keystore = BIP32_KeyStore({})
         keystore.add_seed(seed)
         keystore.passphrase = passphrase
         bip32_seed = Mnemonic.mnemonic_to_seed(seed, passphrase)
-        keystore.add_xprv_from_seed(bip32_seed, "m/")
+        if t == 'standard':
+            der = "m/"
+            xtype = 'standard'
+        else:
+            der = "m/1'/" if is_p2sh else "m/0'/"
+            xtype = 'p2wsh' if is_p2sh else 'p2wpkh'
+        keystore.add_xprv_from_seed(bip32_seed, xtype, der)
+    else:
+        raise BitcoinException('Unexpected seed type {}'.format(t))
     return keystore
 
 def from_private_key_list(text):
     keystore = Imported_KeyStore({})
-    for x in text.split():
+    for x in get_private_keys(text):
         keystore.import_key(x, None)
     return keystore
 
@@ -710,15 +756,13 @@ def from_xprv(xprv):
     k.xpub = xpub
     return k
 
-def from_keys(text):
+def from_master_key(text):
     if is_xprv(text):
         k = from_xprv(text)
     elif is_old_mpk(text):
         k = from_old_mpk(text)
     elif is_xpub(text):
         k = from_xpub(text)
-    elif is_private_key_list(text):
-        k = from_private_key_list(text)
     else:
-        raise BaseException('Invalid key')
+        raise BitcoinException('Invalid master key')
     return k
